var
	/* how many units each cell is divided into.
		Example: 8 units_per_cell chops each cell into 8 x 8 units. */
	units_per_cell = 8
	lookups = 0
	x_step[960]
	y_step[960]
	cos_table[960]
	sin_table[960]
	tan_table[960]
	inv_cos_table[960]
	inv_sin_table[960]
	inv_tan_table[960]


world
	view = "10x10"
	New()
		..()
		spawn()
			world.log << "Initializing lookup tables."
			for(var/X = 1 to 960)
				var/angle = X*3/8
				var/cos = cos(angle)
				var/sin = sin(angle)
				cos_table[X] = cos
				sin_table[X] = sin
				if(sin)
					inv_tan_table[X] = cos/sin
					inv_sin_table[X] = 1/sin
				if(cos)
					tan_table[X] = sin/cos
					inv_cos_table[X] = 1/cos


				if (X >0 && X<480)
					y_step[X] = abs(tan_table[X] * units_per_cell);
				else
					y_step[X] = -abs(tan_table[X] * units_per_cell);
				if (X > 240 && X<= 720)
					x_step[X] = -abs(inv_tan_table[X] * units_per_cell);
				else
					x_step[X] = abs(inv_tan_table[X] * units_per_cell);

				sleep()
			world.log << "Lookup tables complete."
			lookups = 1

turf
	var
		sliver_icon
	New()
		. = ..()
		if(!sliver_icon)
			sliver_icon = 'rt_line.dmi' + icon

client
	var
		show_under = 0
		step_size = 16
		obj
			ceiling = new()
			floor = new()
		list
			vert_line = list()
			line_top = list()
	proc
		RayCaster(x,y,view_angle)
			var
				ray_angle = 880	// used to compensate for fish eye distortion
				quad_y		// quadrant of ray
				quad_x
				x_bound		// cell boundry of next vert and horizontal intersect
				y_bound
				x_delta		// amount to move for next cell position
				y_delta
				cell_x		// world x,y co-ords
				cell_y
				xi			// x intersect
				yi			// y intersect
				x_dist		// distance from viewpoint to intersect
				y_dist
				ray			// current ray being cast
				turf
					x_hit	// records turf intersected
					y_hit
					hit
				icon		// hit.sliver_icon modified by which face is hit

				top			// top of the sliver
				bottom		// bottom of the sliver
				scale		// scale of final sliver

				asymtote	// used to cheat at asymtotes

			// view is 60 degrees wide. Start at 30degrees right of view_angle
			view_angle -= 80	// 80 = 30 deg
			if(view_angle <= 0) view_angle += 960 // 960 = 360 deg

			//loop through 160 rays (one for each vertical strip)
			for(ray = 1 to 160)

			// compute first x intersection
				if(view_angle <= 480) // in upper half plane (480 = 180 deg)
					y_bound = units_per_cell + units_per_cell * (round(y/units_per_cell))
					y_delta = units_per_cell
					xi = inv_tan_table[view_angle] * (y_bound - y) + x
					quad_y = 0
				else // in lower half plane
					y_bound = units_per_cell * (round(y/units_per_cell))
					y_delta = -units_per_cell
					xi = inv_tan_table[view_angle] * (y_bound - y) + x
					quad_y = -1

			// compute first y intersection
				if((view_angle <= 240) || (view_angle > 720)) // in right half plane
					// (240 = 90deg; 720 = 270deg))
					x_bound = units_per_cell + units_per_cell * (round(x/units_per_cell))
					x_delta = units_per_cell
					yi = tan_table[view_angle] * (x_bound - x) + y
					quad_x = 0
				else // left half plane
					x_bound = units_per_cell * (round(x/units_per_cell))
					x_delta = -units_per_cell
					yi = tan_table[view_angle] * (x_bound - x) + y
					quad_x = -1

				// check for y intersects
				x_dist = 0
				while(!x_dist)
					cell_x = round( (x_bound) / units_per_cell) + quad_x
					cell_y = round(yi/ units_per_cell)
					x_hit = locate(cell_x + 1, cell_y + 1, mob.z)
					// check if hit edge of world or visual obstruction
					if(!x_hit || x_hit.opacity)
						x_dist = (yi - y) * inv_sin_table[view_angle]
					else	// advance the ray
						yi += y_step[view_angle]
						x_bound += x_delta

				// check for x intersects
				y_dist = 0
				while(!y_dist)
					cell_x = round(xi/units_per_cell)
					cell_y = round((y_bound)/units_per_cell) + quad_y
					y_hit = locate(cell_x + 1,cell_y + 1, mob.z)
					// check if hit edge of world or visual obstruction
					if(!y_hit || y_hit.opacity)
						y_dist = (xi - x) * inv_cos_table[view_angle]
					else	// advance the ray
						xi += x_step[view_angle]
						y_bound += y_delta

				icon = null
				if(x_dist < y_dist)
					scale = round(inv_cos_table[ray_angle]*400/(0.00000000001 + x_dist))
					hit = x_hit
					// basic turf sliver icon for east and west face
					if(hit) icon = hit.sliver_icon
				else
					scale = round(inv_cos_table[ray_angle]*400/(0.00000000001 + y_dist))
					hit = y_hit
					// lighter version of the sliver icon for north and south face
					if(hit) icon = hit.sliver_icon + 'lighten.dmi'
				top = 80 + scale
				bottom = 80 - scale

				if(hit)	// hit an opaque turf
					VertLine(160-ray,bottom,top,icon,hit.name)
				else	// hit the world border
					VertLine(160-ray,bottom,top,,"world border")

				if(asymtote)	// this angle was an asymtote
					view_angle++	// advance the angle an extra step
					asymtote = 0

				// advanve the view_angle and keep it in 1-960 range
				if(++view_angle > 960)	// 960 = 360deg
					view_angle = 1

				if(!(view_angle%240))	// if the new angle is an asymtote
					asymtote = 1	// flag it
					view_angle--	// use the last angle

				// advanve the ray_angle and keep it in 1-960 range
				if(++ray_angle > 960)	// 960 = 360deg
					ray_angle = 1

		UpdateView()
			// if the mob is on the map and lookup tables are finished
			if(mob.loc && lookups)
				// get micro co-ords of the mob
				var
					x = round((mob.loc.x - 1  + mob.pixel_x / 32) * units_per_cell)
					y = round((mob.loc.y - 1  + mob.pixel_y / 32) * units_per_cell)

				RayCaster(x,y,mob.angle)

		VertLine(x, y1, y2, icon = 'rt_line.dmi', name = "line", layer = 100)
			// draws a vertical line on client screen from x,y1 to x,y2

			// dither to show map beneath
			if(show_under) icon  += 'dither.dmi'

			// enforce limits
			x = max(0,min(159,x))
			y1 = max(0,min(160,y1))
			y2 = max(0,min(160,y2))

			// retrieve line bottom
			var/obj/bottom = vert_line["[x]"]
			//retrive line top
			var/obj/top = line_top["[x]"]

			if(!bottom)	// none cached here, so cache a new line
				bottom = new()
				vert_line["[x]"] = bottom
				screen += bottom
				top = new()
				line_top["[x]"] = top
				screen += top

			bottom.icon = icon
			bottom.layer = layer

			if(y1 >= y2)	// no line to draw
				bottom.icon = null
				top.icon = null
				return

			bottom.name = name
			top.name = name

			// caculate screen_loc co-ords
			var
				Xc = round(x/16) + 1
				Xp = x&15
				Y1c = round(y1/16) + 1
				Y1p = y1&15
				Y2c = round(y2/16)
				Y2p = y2&15
			// double these values since the lines are built of 2x2 blocks
			Xp += Xp
			Y2p += Y2p
			Y1p += Y1p

			if(y2 - y1 < 16)	// tiny line
				top.icon = null	// hide the top
				bottom.icon_state = "[y2-y1]"	// use a partial line icon_state
				bottom.screen_loc = "[Xc]:[Xp],[Y1c]:[Y1p]"
			else
				bottom.icon_state = ""	// use the full line icon_state
				bottom.screen_loc = "[Xc]:[Xp],[Y1c]:[Y1p]"
				top.icon = icon	// make sure top shows
				top.layer = layer
				if(Y2p<Y1p) Y1c++	// move the bottom edge of the line top up
				top.screen_loc = "[Xc]:[Xp],[Y1c]:[Y2p] to [Xc]:[Xp],[Y2c]:[Y2p]"

	Center()	//toggle step size
		step_size += 16
		if(step_size > 32) step_size = 16

	Move(loc,dir)
		var
			dx
			dy
		switch(dir)
			if(NORTH)	//step forward
				dx = step_size * cos_table[mob.angle]
				dy = step_size * sin_table[mob.angle]
			if(SOUTH)	// step back
				dx = -step_size * cos_table[mob.angle]
				dy = -step_size * sin_table[mob.angle]
			if(SOUTHEAST)	// step right
				dy = -step_size * cos_table[mob.angle]
				dx = step_size * sin_table[mob.angle]
			if(SOUTHWEST)	// step left
				dy = step_size * cos_table[mob.angle]
				dx = -step_size * sin_table[mob.angle]
			if(EAST)	// 45 deg turn right
				mob.angle -= 120
				if(mob.angle < 1) mob.angle += 960
				UpdateView()
			if(NORTHEAST)	// 15 deg turn right
				mob.angle -= 40
				if(mob.angle < 1) mob.angle += 960
				UpdateView()
			if(WEST)	// 45 deg turn left
				mob.angle += 120
				if(mob.angle > 960) mob.angle -= 960
				UpdateView()
			if(NORTHWEST)	// 15 deg turn left
				mob.angle += 40
				if(mob.angle > 960) mob.angle -= 960
				UpdateView()

		if(dx || dy)
			var
				nx = mob.x
				ny = mob.y
			dx += mob.pixel_x
			dy += mob.pixel_y
			if(dx < 1)
				nx--
				dx += 32
			if(dx > 32)
				nx++
				dx -= 32
			if(dy < 1)
				ny--
				dy += 32
			if(dy > 32)
				ny++
				dy -= 32

			if((ny != mob.y) || (nx != mob.x))
				mob.Move(locate(nx,ny,mob.z),,dx,dy)
				return
			else
				if(!mob.BorderCheck(mob.loc,dx,dy))
					return 0
				mob.pixel_x = dx
				mob.pixel_y = dy
				UpdateView()

	New()
		. = ..()
		world << "<b>[src] logs in."
		ceiling.icon = 'ceiling.dmi'
		ceiling.screen_loc = "WEST,6 to EAST,NORTH"
		ceiling.layer = 99
		ceiling.name = "ceiling"
		screen += ceiling
		floor.icon = 'floor.dmi'
		floor.screen_loc = "WEST,SOUTH to EAST,5"
		floor.layer = 99
		floor.name = "floor"
		screen += floor
		spawn()
			while(!lookups) sleep(1)
			UpdateView()
	Del()
		world << "<b>[src] logs out."
		del(mob)
		. = ..()

	verb
		say(t as text)
			world << "<b>[src]:</b> [t]"

		map()
			show_under = !show_under
			if(show_under)
				ceiling.icon += 'dither.dmi'
				floor.icon += 'dither.dmi'
			else
				ceiling.icon = 'ceiling.dmi'
				floor.icon = 'floor.dmi'
			UpdateView()

mob
	icon = 'mob.dmi'
	pixel_x = 15
	pixel_y = 15
	animate_movement = 0
	sight = SEE_TURFS
	var
		angle = 960
		width = 4

	Stat()
		stat("loc","[x],[y]")
		stat("pixel","[pixel_x],[pixel_y]")
		stat("angle","[angle*3/8] deg")

	Move(Loc,Dir,dx,dy)
		if(isnum(dx) && !BorderCheck(Loc,dx,dy))
			return 0
		. = ..()
		if(. && client)
			if(isnum(dx))
				pixel_x = dx
				pixel_y = dy
			client.UpdateView()
	proc
		BorderCheck(Loc,dx,dy)
			// check if the mob is too close to something
			if(dx - width <= 0)
				var/turf/T = get_step(Loc,WEST)
				if(!T || !T.Enter(src))
					return 0
			else if(dx + width >= 32)
				var/turf/T = get_step(Loc,EAST)
				if(!T || !T.Enter(src))
					return 0
			if(dy - width <= 0)
				var/turf/T = get_step(Loc,SOUTH)
				if(!T || !T.Enter(src))
					return 0
			else if(dy + width >= 32)
				var/turf/T = get_step(Loc,NORTH)
				if(!T || !T.Enter(src))
					return 0
			return 1